/*
                LFDirStack V 1.1
                  LF soft 1992

    A directory stack for Amiga . Use Arp.library V39+.

    because Csh 4.02A don't have this usefull stuff ...

        LFDirStack is a freeware but write me if you want to include it
    on commercial product. It was compiled /w  FreeWare version of Dice
    ( M.Dillon), tested /w ARP.library V39 on a 68010, 2Mb A1000.

    Send bugs reports , hints or ... money to

                        Laurent Faillie
                        " Les Vuardes "
                        74930 Pers-Jussy
                            FRANCE
                                                & sorry for my poor english
*/

#include <stdlib.h>
#include <stdio.h>
#define FastLF
#include <LF.h>
#include <exec/memory.h>
#include <libraries/dos.h>
#include <libraries/dosextens.h>

#define SDIRLEN 120

UBYTE *stackbase=NULL;

const char  *LFStackPtr="LFStackPtr",
            *err_arg="Illegal argument",
            *err_nostack="No stack",
            *err_est="Stack empty";

void Erreur(const char *msg){
    printf("[33m%s[0m\n",msg);
    exit(15);
}

struct PERROR {
        const int errnum;                     /* Format of global error lookup */
        const char *errstr;
};

const struct PERROR Perror[]= {       /* error code->string */
        103,    "Insufficient free storage",
        105,    "Task table full",
        120,    "Argument line invalid or too long",
        121,    "File is not an object module",
        122,    "Invalid resident library during load",
        201,    "No default directory",
        202,    "Object in use",
        203,    "Object already exists",
        204,    "Directory not found",
        205,    "Object not found",
        206,    "Bad stream name",
        207,    "Object too large",
        209,    "Action not known",
        210,    "Invalid stream component name",
        211,    "Invalid object lock",
        212,    "Object not of required type",
        213,    "Disk not validated",
        214,    "Disk write protected",
        215,    "Rename across devices",
        216,    "Directory not empty",
        217,    "Too many levels",
        218,    "Device not mounted",
        219,    "Seek error",
        220,    "Comment too long",
        221,    "Disk full",
        222,    "File delete protected",
        223,    "File write protected",
        224,    "File read protected",
        225,    "Not a DOS disk",
        226,    "No disk",
        0,      NULL
};

void ierror(str, err)
register char *str;
{
   register struct PERROR *per = Perror;

#ifdef DEBUG
    puts("*D* ierror");
#endif

   if (err) {
      for (; per->errstr; ++per) {
         if (per->errnum == err) {
            printf("[33m%s%s%s[0m\n",
                  per->errstr,
                  (str) ? ": " : "",
                  (str) ? str : "");
            exit(15);
         }
      }
      printf ("[33mUnknown DOS error %d %s[0m\n", err, (str) ? str : "");
   }
   exit(15);
}

void pError(str)
char *str;
{
int ierr = (long)IoErr();
ierror(str, ierr);
}

void help()
{
#ifdef DEBUG
    puts("*D* help");
#endif
    puts("\n\n          [33;42;1m LFDirStack V 1.1 [0m");
    puts("            [33;42m  LFsoft 1992 [0m");
    puts("\n\nLFDirStack -i [nbre] : Init Stack,allocate memory for <nbre> dirname (def 20)");
    puts("LFDirStack -ph [dir] : Push current directory or dir on stack");
    puts("LFDirStack -pp       : Pop top of stack");
    puts("LFDirStack -l        : List stack");
    puts("LFDirStack -s [num]  : set cd with top of stack or dir num on stack");
    puts("LFDirStack -c        : Clear stack & free memory\n\n");
}

char verify()
{
#ifdef DEBUG
    puts("*D* verify");
#endif
    if(stackbase){
        if(stackbase[1]!=SDIRLEN || !*stackbase)
            Erreur("Fatal error: strange LFStackPtr found");
        else return(1);
    } else return(0);
}

UBYTE *stel(int num)
{
    return (stackbase+2+stackbase[1]*num);
}

void init(unsigned char nbre)
{
#ifdef DEBUG
    puts("*D* init");
#endif

    if(stackbase)
        Erreur("Stack already in use");
    if(!nbre)
        Erreur(err_arg);

    if(!(stackbase=(char *)AllocMem(2+nbre*SDIRLEN,MEMF_PUBLIC|MEMF_CLEAR)))
        Erreur("Can't allocate stack memory");
    else{
        char tmp[256];
        sprintf(tmp,"%ld",stackbase);
        Setenv(LFStackPtr,tmp);
        *stackbase=nbre;
        stackbase[1]=SDIRLEN;
    }
}

void push(char *s)
{
    int i;
#ifdef DEBUG
    puts("*D* push");
#endif

    if(!verify())
        Erreur (err_nostack);

    for(i=0;i<*stackbase;i++)
        if (!*stel(i)){
            strcpy(stel(i),s);
            return();
        }

    Erreur("Stack full");
}

void pop()
{
    BPTR oldlock, filelock;
    char *str;
    int i;

#ifdef DEBUG
    puts("*D* pop");
#endif

    if(!verify())
        Erreur (err_nostack);

    for(i=0;i<*stackbase && *stel(i);i++);

#ifdef DEBUG
    printf("1er libre :%ld\n",i);
#endif

    if(!i--)
        Erreur (err_est);

    filelock=Lock((str=stel(i)),ACCESS_READ);
    if (filelock==NULL) { pError(str); return 20; }
    if (oldlock=CurrentDir(filelock)) UnLock(oldlock);

    *str=0;
}

void set(int num)
{
    BPTR oldlock, filelock;
    char *str;
    int i;

#ifdef DEBUG
    puts("*D* pop");
#endif

    if(!verify())
        Erreur (err_nostack);

    for(i=0;i<*stackbase && *stel(i);i++);

#ifdef DEBUG
    printf("1er libre :%ld\n",i);
#endif

    if(!i--)
        Erreur (err_est);

    if(num==-1)
        num=i;

    if(num>i)
        Erreur ("Undefined element");

    filelock=Lock((str=stel(num)),ACCESS_READ);
    if (filelock==NULL) { pError(str); return 20; }
    if (oldlock=CurrentDir(filelock)) UnLock(oldlock);
}

void list()
{
#ifdef DEBUG
    puts("*D* list");
#endif
    if(verify()){
        int i;

        printf("[3m%ld dirs reserved[0m\n",*stackbase);
        for (i=0;i<*stackbase;i++)
            if(*stel(i))
                printf("[1;2m%ld: [0m%s\n",i,stel(i));
            else
                break;
    }else
        Erreur(err_nostack);
}

void fst()
{
#ifdef DEBUG
    puts("*D* fst");
#endif
    if(verify()){
        FreeMem(stackbase,2+(*stackbase)*stackbase[1]);
        Setenv(LFStackPtr,"0");
    }else
        Erreur(err_nostack);
}

void main(ac,av)
int ac;
char **av;
{
    char tmp[256];

    OpenArp();
    atexit(CloseArp);

    if(ac==1){
        help();
        exit(0);
    }

    if (Getenv(LFStackPtr,tmp,10L)) stackbase=(void *)atol(tmp);
#ifdef DEBUG
    printf("*D* stackbase = %lx (%s)\n",stackbase,tmp);
#endif

    if(!strcmp(av[1],"-i")){
        if(ac >3)
            Erreur(err_arg);
        else if (ac==3)
            init(atoi(av[2]));
        else init(20);
    }else if(!strcmp(av[1],"-ph")){
        if(ac==3)
            push(av[2]);
        else {
            getcwd(tmp,119);
            push(tmp);
        }
    }else if(!strcmp(av[1],"-pp"))
        pop();
    else if(!strcmp(av[1],"-l"))
        list();
    else if(!strcmp(av[1],"-s"))
        if(ac==3)
            set(atoi(av[2]));
        else
            set(-1);
    else if(!strcmp(av[1],"-f"))
        fst();
    else help();
    exit(0);
}
